
#include "kyokumen.h"
#include "move_list.h"

#include <memory.h>
#include <stdlib.h>
#include <algorithm> // min(), max(), swap()
using namespace std;

uint64 Kyokumen::HashSeed[ERY+1][0x99+1];
uint64 Kyokumen::HandHashSeed[EHI+1][18+1];

uint64 Kyokumen::HashHistory[1000];
int Kyokumen::OuteHistory[1000];

void Kyokumen::HashInit()
{
  // ISSUE: bit or しているので、ほとんどのビットか1になってしまう。
  int i;
  for(i=0;i<=ERY;i++) {
    for(int pos=0x11;pos<=0x99;pos++) {
			HashSeed[i][pos]=((uint64)rand())<<49|
							 ((uint64)rand())<<34|
							 ((uint64)rand())<<19|
							 ((uint64)rand())<<4|
							  rand() & 0x07;
		}
	}
	
  for(i=SFU;i<=EHI;i++) {
    for(int maisuu=0;maisuu<=18;maisuu++) {
      HandHashSeed[i][maisuu]=((uint64)rand())<<49|
	((uint64)rand())<<34|
	((uint64)rand())<<19|
	((uint64)rand())<<4|
	rand() & 0x07;
    }
  }
}



/** 局面のコンストラクタ：盤面の状態と手数、持ち駒から生成.
 */
Kyokumen::Kyokumen(int tesu, const KomaInf board[9][9], const int Motigoma[])
{
  // 盤面をWALL（壁）で埋めておきます。
  memset(banpadding,WALL,sizeof(banpadding));
  memset(ban,WALL,sizeof(ban));

  // 初期化
  value=0;

  kingS=0;
  kingE=0;
  Tesu=tesu;

  HashVal=0;
  HandHashVal=0;
  KyokumenHashVal=0;

  // boardで与えられた局面を設定します。
  for(int dan=1;dan<=9;dan++) {
    for(int suji=0x10;suji<=0x90;suji+=0x10) {
      // 将棋の筋は左から右なので、配列の宣言と逆になるため、筋はひっくり返さないとなりません。
      ban[suji+dan]=board[dan-1][9-suji/0x10];
      KyokumenHashVal^=HashSeed[ban[suji+dan]][suji+dan];
      if (ban[suji+dan]==SOU)
	kingS = suji + dan;
      else if (ban[suji+dan]==EOU)
	kingE = suji + dan;

      value+=KomaValue[ban[suji+dan]];
    }
  }

  // 持ち駒はそのまま利用します。
  int i;
  for(i = 0; i <= EHI; i++) {
    Hand[i] = Motigoma[i];
    value += HandValue[i] * Hand[i];
    for(int j = 1; j <= Hand[i];j++) {
      HandHashVal ^= HandHashSeed[i][j];
    }
  }

  HashVal = KyokumenHashVal ^ HandHashVal;
  for(i=0;i<Tesu;i++) {
    HashHistory[i]=0;
    OuteHistory[i]=0;
  }
  HashHistory[Tesu]=HashVal;
  OuteHistory[Tesu]=((Tesu%2)==0)?controlS[kingE]:controlE[kingS];

  // controlS/controlEを初期化します。
  InitControl();
}


// controlS,controlEの初期化
void Kyokumen::InitControl()
{
  int dan, suji;
  int i, j, b, bj;

  memset(controlS,0,sizeof(controlS));
  memset(controlE,0,sizeof(controlE));

  for (suji = 0x10; suji <= 0x90; suji += 0x10) {
    for (dan = 1 ; dan <= 9 ; dan++) {
      if (ban[suji + dan] & ENEMY) { //敵の駒
        //駒の効きを追加する
	for (i = 0, b = 1, bj = (1<<16); i < 12; i++, b<<=1, bj<<=1) {
	  if (CanJump[i][ban[dan + suji]]) {
	    j = dan + suji;
	    do {
	      j += Direct[i];
	      controlE[j] |= bj;
	    } while (ban[j] == EMPTY);
	  } else if (CanMove[i][ban[dan + suji]]) {
	    controlE[dan + suji + Direct[i]] |= b;
	  }
	}
      }
      else if (ban[suji + dan] & SELF) { //味方の駒が有る
        //駒の効きを追加する
	for (i = 0, b = 1, bj = (1<<16); i < 12; i++, b<<=1, bj<<=1) {
	  if (CanJump[i][ban[dan + suji]]) {
	    j = dan + suji;
	    do {
	      j += Direct[i];
	      controlS[j] |= bj;
	    } while (ban[j] == EMPTY);
	  } else if (CanMove[i][ban[dan + suji]]) {
	    controlS[dan + suji + Direct[i]] |= b;
	  }
	}
      }	
    }
  }
}


/** 手で局面を進める.
 */
void Kyokumen::Move(int SorE, const Te& te)
{
  int i,j,b,bj;
  if (te.from > 0x10) {
    // 元いた駒のコントロール (利き) を消す
    int dir;
    for(dir=0,b=1,bj=1<<16;dir<12;dir++,b<<=1,bj<<=1) {
      if (SorE==SELF) {
	controlS[te.from+Direct[dir]]&=~b;
      } else {
	controlE[te.from+Direct[dir]]&=~b;
      }
      if (CanJump[dir][te.koma]) {
	int j = te.from;
	do {
	  j += Direct[dir];
	  if (SorE==SELF) {
	    controlS[j] &= ~bj;
	  } else {
	    controlE[j] &= ~bj;
	  }
	} while (ban[j] == EMPTY);
      }
    }

    // 元いた位置は空白になる
    ban[te.from] = EMPTY;

    // 空白になったことで変わるハッシュ値
    KyokumenHashVal^=HashSeed[te.koma][te.from];
    KyokumenHashVal^=HashSeed[EMPTY][te.from];
    
    // 飛び利きを伸ばす
    for (i = 0, bj = (1<<16); i < 8; i++, bj<<=1) {
      int Dir=Direct[i];
      if (controlS[te.from] & bj) {
	j = te.from;
	do {
	  j += Dir;
	  controlS[j] |= bj;
	} while (ban[j] == EMPTY);
      }
      if (controlE[te.from] & bj) {
	j = te.from;
	do {
	  j += Dir;
	  controlE[j] |= bj;
	} while (ban[j] == EMPTY);
      }
    }
  } 
  else {
    // 持ち駒から一枚減らす
    HandHashVal ^= HandHashSeed[te.koma][Hand[te.koma]];
    Hand[te.koma]--;
    value-=HandValue[te.koma];
    value+=KomaValue[te.koma];
  }

  if (ban[te.to] != EMPTY) {
    // 相手の駒を持ち駒にする。
    // 持ち駒にする時は、成っている駒も不成りに戻す。（&~PROMOTED）
    value-=KomaValue[ban[te.to]];
    value+=HandValue[SorE|(ban[te.to]&~PROMOTED&~SELF&~ENEMY)];
    int koma=SorE|(ban[te.to]&~PROMOTED&~SELF&~ENEMY);
    Hand[koma]++;
    // ハッシュに取った駒を加える
    HandHashVal^=HandHashSeed[koma][Hand[koma]];

    //取った駒の効きを消す
    for (i = 0, b = 1, bj = (1<<16); i < 12; i++, b<<=1, bj<<=1) {
      int Dir=Direct[i];
      if (CanJump[i][ban[te.to]]) {
	j = te.to;
	do {
	  j += Dir;
	  if (SorE==SELF) {
	    controlE[j] &= ~bj;
	  } else {
	    controlS[j] &= ~bj;
	  }
	} while (ban[j] == EMPTY);
      } else {
	j=te.to + Dir;
	if (SorE==SELF) {
	  controlE[j] &= ~b;
	} else {
	  controlS[j] &= ~b;
	}
      }
    }
  } 
  else {
    // 移動先で遮った飛び利きを消す
    for (i = 0, bj = (1<<16); i < 8; i++, bj<<=1) {
      int Dir=Direct[i];
      if (controlS[te.to] & bj) {
	j = te.to;
	do {
	  j += Dir;
	  controlS[j] &= ~bj;
	} while (ban[j] == EMPTY);
      }
      if (controlE[te.to] & bj) {
	j = te.to;
	do {
	  j += Dir;
	  controlE[j] &= ~bj;
	} while (ban[j] == EMPTY);
      }
    }
  }
  // ban[te.to]にあったものをＨａｓｈから消す
  KyokumenHashVal ^= HashSeed[ban[te.to]][te.to];
  if (te.promote) {
    value-=KomaValue[te.koma];
    value+=KomaValue[te.koma|PROMOTED];
    ban[te.to]=te.koma|PROMOTED;
  } 
  else {
    ban[te.to] = te.koma;
  }

  // 新しい駒をＨａｓｈに加える
  KyokumenHashVal ^= HashSeed[ban[te.to]][te.to];

  // 移動先の利きをつける
  for (i = 0, b = 1, bj = (1<<16); i < 12; i++, b<<=1, bj<<=1) {
    if (CanJump[i][ban[te.to]]) {
      j = te.to;
      do {
	j += Direct[i];
	if (SorE==SELF) {
	  controlS[j] |= bj;
	} else {
	  controlE[j] |= bj;
	}
      } while (ban[j] == EMPTY);
    } 
    else if (CanMove[i][ban[te.to]]) {
      if (SorE==SELF) {
	controlS[te.to+Direct[i]] |= b;
      } else {
	controlE[te.to+Direct[i]] |= b;
      }
    }
  }

  // 玉の位置は覚えておく。
  if (te.koma == SOU) 
    kingS = te.to;
  else if (te.koma == EOU) 
    kingE = te.to;

  HashVal = KyokumenHashVal ^ HandHashVal;
  Tesu++;

  HashHistory[Tesu] = HashVal;
  OuteHistory[Tesu] = (SorE == SELF) ? controlS[kingE]:controlE[kingS];
}


/** ピン（動かすと王を取られてしまうので動きが制限される駒）の状態を設定する.
 */
void Kyokumen::MakePinInf(int *pin) const
{
  int i;
  // ピン情報を設定する
  for (i = 0x11; i <= 0x99; i++) {
    // 0はピンされていない、という意味
    pin[i] = 0;
  }

  if (kingS) {	//自玉が盤面にある時のみ有効
    for (i = 0; i < 8; i++) {
      int p;
      p = search(kingS, -Direct[i]); 
      if ((ban[p] != WALL) && !(ban[p] & ENEMY)) { //味方の駒が有る
        if (controlE[p] & (1<<(16+i))) {
          pin[p] = Direct[i];
        }
      }
    }
  }

  if (kingE) {	//敵玉が盤面にある時のみ有効
    for (i = 0; i < 8; i++) {
      int p;
      p = search(kingE, -Direct[i]);
      if ((ban[p] != WALL) && (ban[p] & ENEMY)) { //敵の駒が有る
	if (controlS[p]&(1<<(16+i))) {
	  pin[p]=Direct[i];
	}
      }
    }
  }
}


/** 駒の動きとして正しい動きを全て生成する.
    @param SorE   手番
 */
int Kyokumen::MakeLegalMoves(int SorE, Te* teBuf, int* pin)
{
  int teNum = 0;

  int pbuf[16*11];
  if (pin==NULL) {
    MakePinInf(pbuf);
    pin=pbuf;
  }
  if (SorE == SELF && controlE[kingS]!=0) {
    return AntiCheck(SorE,teBuf,pin,controlE[kingS]);
  }
  if (SorE==ENEMY && controlS[kingE] != 0) {
    return AntiCheck(SorE, teBuf, pin, controlS[kingE]);
  }

  int suji,dan;
  int StartDan,EndDan;

  // 盤上の駒を動かす
  for(suji=0x10;suji<=0x90;suji+=0x10) {
    for(dan=1;dan<=9;dan++) {
      if (ban[suji+dan]&SorE) {
	AddMoves(SorE,teNum,teBuf,suji+dan,pin[suji+dan]);
      }
    }
  }

  // 歩を打つ
  if (Hand[SorE|FU]>0) {
    for(suji=0x10;suji<=0x90;suji+=0x10) {
      // 二歩チェック
      int nifu=0;
      for(dan=1;dan<=9;dan++) {
	if (ban[suji+dan]==(SorE|FU)) {
	  nifu=true;
	  break;
	}
      }
      if (nifu) continue;
      //(先手なら２段目より下に、後手なら８段目より上に打つ）
      if (SorE==SELF) {
	StartDan=2;
	EndDan=9;
      } else {
	StartDan=1;
	EndDan=8;
      }
      for(dan=StartDan;dan<=EndDan;dan++) {
	// 打ち歩詰めもチェック
	if (ban[dan+suji]==EMPTY && !Utifudume(SorE,dan+suji,pin)) {
	  teBuf[teNum++]=Te(0,suji+dan,SorE|FU,EMPTY);
	}
      }
    }
  }
  // 香を打つ
  if (Hand[SorE|KY]>0) {
    for(suji=0x10;suji<=0x90;suji+=0x10) {
      //(先手なら２段目より下に、後手なら８段目より上に打つ）
      if (SorE==SELF) {
	StartDan=2;
	EndDan=9;
      } else {
	StartDan=1;
	EndDan=8;
      }
      for(dan=StartDan;dan<=EndDan;dan++) {
	if (ban[dan+suji]==EMPTY) {
	  teBuf[teNum++]=Te(0,suji+dan,SorE|KY,EMPTY);
	}
      }
    }
  }
  //桂を打つ
  if (Hand[SorE|KE]>0) {
    //(先手なら３段目より下に、後手なら７段目より上に打つ）
    for(suji=0x10;suji<=0x90;suji+=0x10) {
      if (SorE==SELF) {
	StartDan=3;
	EndDan=9;
      } else {
	StartDan=1;
	EndDan=7;
      }
      for(dan=StartDan;dan<=EndDan;dan++) {
	if (ban[dan+suji]==EMPTY) {
	  teBuf[teNum++]=Te(0,suji+dan,SorE|KE,EMPTY);
	}
      }
    }
  }
  // 銀〜飛車は、どこにでも打てる
  for (int koma = GI; koma <= HI; koma++) {
    if (Hand[SorE | koma] > 0) {
      for(suji=0x10;suji<=0x90;suji+=0x10) {
	for(dan=1;dan<=9;dan++) {
	  if (ban[dan+suji]==EMPTY) {
	    teBuf[teNum++]=Te(0,suji+dan,SorE|koma,EMPTY);
	  }
	}
      }
    }
  }
	
  return teNum;
}


/** 盤面のfromにある駒を動かす手を生成する.
    MakeLegalMoves() だけから呼び出される。
 */
void Kyokumen::AddMoves(int SorE,int& teNum, Te* teTop,
			int from, int pin,int Rpin)
{
  switch(ban[from]) {
  case SFU:
    AddMove(SorE,teNum,teTop,from,-1,pin,Rpin);
    break;
  case EFU:
    AddMove(SorE,teNum,teTop,from,+1,pin,Rpin);
    break;
  case SKY:
    AddStraight(SorE,teNum,teTop,from,-1,pin,Rpin);
    break;
  case EKY:
    AddStraight(SorE,teNum,teTop,from,+1,pin,Rpin);
    break;
  case SKE:
    AddMove(SorE,teNum,teTop,from,+14,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,-18,pin,Rpin);
    break;
  case EKE:
    AddMove(SorE,teNum,teTop,from,-14,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,+18,pin,Rpin);
    break;
  case SGI:
    AddMove(SorE,teNum,teTop,from, -1,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,-17,pin,Rpin);
    AddMove(SorE,teNum,teTop,from, 15,pin,Rpin);
    AddMove(SorE,teNum,teTop,from, 17,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,-15,pin,Rpin);
    break;
  case EGI:
    AddMove(SorE,teNum,teTop,from, +1,pin,Rpin);
    AddMove(SorE,teNum,teTop,from, 17,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,-15,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,-17,pin,Rpin);
    AddMove(SorE,teNum,teTop,from, 15,pin,Rpin);
    break;
  case SKI:case STO:case SNY:case SNK:case SNG:
    AddMove(SorE,teNum,teTop,from, -1,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,-17,pin,Rpin);
    AddMove(SorE,teNum,teTop,from, 15,pin,Rpin);
    AddMove(SorE,teNum,teTop,from, +1,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,-16,pin,Rpin);
    AddMove(SorE,teNum,teTop,from, 16,pin,Rpin);
    break;
  case EKI:case ETO:case ENY:case ENK:case ENG:
    AddMove(SorE,teNum,teTop,from, +1,pin,Rpin);
    AddMove(SorE,teNum,teTop,from, 17,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,-15,pin,Rpin);
    AddMove(SorE,teNum,teTop,from, -1,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,-16,pin,Rpin);
    AddMove(SorE,teNum,teTop,from, 16,pin,Rpin);
    break;
  case SRY:case ERY:
    AddMove(SorE,teNum,teTop,from, 17,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,-15,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,-17,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,+15,pin,Rpin);
  case SHI:case EHI:
    AddStraight(SorE,teNum,teTop,from,+1,pin,Rpin);
    AddStraight(SorE,teNum,teTop,from,-1,pin,Rpin);
    AddStraight(SorE,teNum,teTop,from,-16,pin,Rpin);
    AddStraight(SorE,teNum,teTop,from,+16,pin,Rpin);
    break;
  case SUM:case EUM:
    AddMove(SorE,teNum,teTop,from,+1,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,+16,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,-16,pin,Rpin);
    AddMove(SorE,teNum,teTop,from,-1,pin,Rpin);
  case SKA:case EKA:
    AddStraight(SorE,teNum,teTop,from,+17,pin,Rpin);
    AddStraight(SorE,teNum,teTop,from,-17,pin,Rpin);
    AddStraight(SorE,teNum,teTop,from,+15,pin,Rpin);
    AddStraight(SorE,teNum,teTop,from,-15,pin,Rpin);
    break;
  case SOU:case EOU:
    MoveKing(SorE,teNum,teTop,0);	// 王手がかかっている時には、AntiCheckの方が呼ばれるから、Kikiは０です。
  }
}


// ある場所の利き情報を作成して返す。普段は使わない関数（差分計算しているから）だが、
// 打ち歩詰めのチェックなど、駒を仮に置いてみて何かするようなときに使用する。
Kiki Kyokumen::CountControlS(int pos) const
{
  Kiki ret=0;
  int i,b,bj;
  b=1;
  bj=1<<16;
  for (i=0; i < DIRECTION_SIZE; i++,b<<=1,bj<<=1) {
    if (CanMove[i][ban[pos-Direct[i]]] && (ban[pos-Direct[i]]&SELF)) {
      ret|=b;
    } else if (CanJump[i][ban[search(pos,-Direct[i])]] && (ban[search(pos,-Direct[i])]&SELF)) {
      ret|=bj;
    }
  }
  return ret;
}


// ある場所の利き情報を作成して返す。普段は使わない関数（差分計算しているから）だが、
// 打ち歩詰めのチェックなど、駒を仮に置いてみて何かするようなときに使用する。
Kiki Kyokumen::CountControlE(int pos) const
{
  Kiki ret=0;
  int i,b,bj;
  b=1;
  bj=1<<16;
  for(i=0;i<12;i++,b<<=1,bj<<=1) {
    if (CanMove[i][ban[pos-Direct[i]]] && (ban[pos-Direct[i]]&ENEMY)) {
      ret|=b;
    } else if (CanJump[i][ban[search(pos,-Direct[i])]] && (ban[search(pos,-Direct[i])]&ENEMY)) {
      ret|=bj;
    }
  }
  return ret;
}


// ある場所に移動できる駒を全部集めて、Kiki情報にして返す。
// このとき、pinされている駒はpinの方向にしか動けない。
Kiki Kyokumen::CountMove(int SorE,int pos,int *pin)
{
  Kiki ret=0;
  int i,b,bj;
  b=1;
  bj=1<<16;
  for(i=0;i<12;i++,b<<=1,bj<<=1) {
    if (CanMove[i][ban[pos-Direct[i]]] && (ban[pos-Direct[i]]&SorE) && (pin[pos-Direct[i]]==0||pin[pos-Direct[i]]==Direct[i]||pin[pos-Direct[i]]==-Direct[i])) {
      ret|=b;
    } 
    else if (CanJump[i][ban[search(pos,-Direct[i])]] && (ban[search(pos,-Direct[i])]&SorE)&& (pin[search(pos,-Direct[i])]==0||pin[search(pos,-Direct[i])]==Direct[i]||pin[search(pos,-Direct[i])]==-Direct[i])) {
      ret|=bj;
    }
  }
  return ret;
}


/** 打ち歩詰めの判定
    MakeLegalMoves(), PutTo() などから呼び出される。
 */
int Kyokumen::Utifudume(int SorE, int to, int *pin)
{
  if (SorE == SELF) {
    // まず、玉の頭に歩を打つ手じゃなければ打ち歩詰めの心配はない。
    if (kingE + 1 != to) 
      return 0;
  } 
  else {
    // まず、玉の頭に歩を打つ手じゃなければ打ち歩詰めの心配はない。
    if (kingS - 1 != to)
      return 0;
  }

  //実際に歩を打って確かめてみる。
  ban[to] = FU | SorE;
  if (SorE == SELF) {
    // 自分の利きがあったら相手は玉で取れない　＆　取る動きを列挙してみたら玉で取る手しかない
    if (controlS[to] && (CountMove(ENEMY,to,pin)==1<<1)) {
      // 玉に逃げ道があるかどうかをチェック
      for (int i = 0; i < 8; i++) {
	KomaInf koma=ban[kingE + Direct[i]];
	if (!(koma & ENEMY) && !CountControlS(kingE+Direct[i])) {
	  // 逃げ道があったので、盤面を元の状態に戻して、
	  ban[to]=EMPTY;
	  // 打ち歩詰めではなかった。
	  return 0;
	}
      }
      // 玉の逃げ道もないのなら、打ち歩詰め。盤面の状態は元に戻す。
      ban[to] = EMPTY;
      return 1;
    }
    // 玉以外で取れる手があるか、玉で取れる。
    ban[to] = EMPTY;
    return 0;
  }
  else {
    // 自分の利きがあったら相手は玉で取れない　＆　取る動きを列挙してみたら玉で取る手しかない
    if (controlE[to] && (CountMove(SELF,to,pin)==1<<6)) {
      // 玉に逃げ道があるかどうかをチェック
      for (int i = 0; i < 8; i++) {
	KomaInf koma=ban[kingS + Direct[i]];
	if (!(koma & SELF) && !CountControlE(kingS+Direct[i])) {
	  // 逃げ道があったので、盤面を元の状態に戻して、
	  ban[to]=EMPTY;
	  // 打ち歩詰めではなかった。
	  return 0;
	}
      }
      // 玉の逃げ道もないのなら、打ち歩詰め。盤面の状態は元に戻す。
      ban[to]=EMPTY;
      return 1;
    }
    // 玉以外で取れる手があるか、玉で取れる。
    ban[to] = EMPTY;
    return 0;
  }
}


/** ある場所（to）に駒を打つ手の生成
    AntiCheck() だけから呼び出される。
 */
void Kyokumen::PutTo(int SorE, int& teNum, Te *teTop, int to, int *pin)
{
  int dan=to &0x0f;
  if (SorE==ENEMY) {
    dan=10-dan;
  }
  if (Hand[SorE|FU]>0 && dan>1) {
    // 歩を打つ手を生成
    // 二歩チェック
    int suji=to & 0xf0;
    int nifu=0;
    for(int d=1;d<=9;d++) {
      if (ban[suji+d]==(SorE|FU)) {
	nifu=1;
	break;
      }
    }
    // 打ち歩詰めもチェック
    if (!nifu && !Utifudume(SorE,to,pin)) {
      teTop[teNum++]=Te(0,to,SorE|FU,EMPTY);
    }
  }
  if (Hand[SorE|KY]>0 && dan>1) {
    // 香を打つ手を生成
    teTop[teNum++]=Te(0,to,SorE|KY,EMPTY);
  }
  if (Hand[SorE|KE]>0 && dan>2) {
    teTop[teNum++]=Te(0,to,SorE|KE,EMPTY);
  }
  for(int koma=GI;koma<=HI;koma++) {
    if (Hand[SorE|koma]>0) {
      teTop[teNum++]=Te(0,to,SorE|koma,EMPTY);
    }
  }
}


/** 王手を受ける手を生成する
 */
int Kyokumen::AntiCheck(int SorE, Te* teBuf, int *pin,Kiki kiki)
{
  int king;
  int teNum=0;
  if ((kiki & (kiki - 1)) != 0) {
    //両王手は玉を動かすしかない
    MoveKing(SorE, teNum, teBuf, kiki);
  }
  else {
    if (SorE == SELF) {
      king = kingS;
    } else {
      king = kingE;
    }

    unsigned int id;
    int check;

    for (id = 0; id <= 31; id++) {
      if (kiki == (1u << id)) break;
    }
    if (id < 16) {
      check = king - Direct[id];
    } else {
      check = search(king,-Direct[id-16]);
    }
    
    //王手駒を取る
    MoveTo(SorE,teNum, teBuf, check, pin);
		
    //玉を動かす
    MoveKing(SorE,teNum, teBuf, kiki);

    if (id >= 16) {
      //合駒をする手を生成する
      int i;
      for (i = king - Direct[id-16]; ban[i] == EMPTY; i -= Direct[id-16]) {
	MoveTo(SorE,teNum, teBuf, i, pin); //移動合
      }
      for (i = king - Direct[id-16]; ban[i] == EMPTY; i -= Direct[id-16]) {
	PutTo(SorE,teNum, teBuf, i, pin);  //駒を打つ合
      }
    } 
  }
  return teNum;
}


/** 玉を動かす手の生成.
    普通の駒と違い、相手の利きのあるところには動けないので、そのための特殊な処理をしています。
 */
void Kyokumen::MoveKing(int SorE, int& teNum, Te *teTop,Kiki kiki)
{
  int i;
  int id = -1;	//隣接王手駒の位置のid
  // 両王手でないなら王手駒の位置を探す
  for (i = 0; i < 8; i++) {
    if (kiki & (1 << i)) {
      id = i;
      break;
    }
  }
  if (id >= 0) {
    // 隣接の王手 最初に取る手を生成するのだ
    if (SorE==SELF) {
      KomaInf koma=ban[kingS-Direct[id]];
      if (( koma==EMPTY || (koma & ENEMY))
	  && !CountControlE(kingS - Direct[id]) //敵の駒が効いていない
	  && !(kiki & (1 << (23-id))))  //敵の飛駒で貫かれていない
	AddMove(SorE,teNum, teTop, kingS, -Direct[id], 0);
    } 
    else {
      KomaInf koma=ban[kingE-Direct[id]];
      if (( koma==EMPTY || (koma & SELF))
	  && !CountControlS(kingE - Direct[id]) //敵の駒が効いていない
	  && !(kiki & (1 << (23-id))))  //敵の飛駒で貫かれていない
	AddMove(SorE,teNum, teTop, kingE, -Direct[id], 0);
    }
  }
  for (i = 0; i < 8; i++) {
    if (i == id) continue;
    if (SorE==SELF) {
      KomaInf koma=ban[kingS-Direct[i]];
      if (( koma==EMPTY || (koma & ENEMY))
	  && !CountControlE(kingS - Direct[i]) //敵の駒が効いていない
	  && !(kiki & (1 << (23-i))))  //敵の飛駒で貫かれていない
	AddMove(SorE,teNum, teTop, kingS, -Direct[i], 0);
    } 
    else {
      KomaInf koma=ban[kingE-Direct[i]];
      if (( koma==EMPTY || (koma & SELF))
	  && !CountControlS(kingE - Direct[i]) //敵の駒が効いていない
	  && !(kiki & (1 << (23-i))))  //敵の飛駒で貫かれていない
	AddMove(SorE, teNum, teTop, kingE, -Direct[i], 0);
    }
  }
}


/** 手の生成：成り・不成りも意識して、駒の動く手を生成する。
 */
void Kyokumen::AddMove(int SorE, int& teNum, Te *teTop,
		       int from, int diff, int pin,int Rpin)
{
  if (Rpin==diff||Rpin==-diff) {
    return;
  }
  int to=from+diff;
  int dan=to&0x0f;
  int fromDan=from&0x0f;

  if ((pin==0 || pin==diff || pin==-diff) && !(ban[to]&SorE)) {
    if (ban[from]==SKE && dan<=2) {
      // 必ず成る
      teTop[teNum++]=Te(from,to,ban[from],ban[to],1);
    } 
    else if ((ban[from]==SFU || ban[from]==SKY) && dan<=1) {
      // 必ず成る
      teTop[teNum++]=Te(from,to,ban[from],ban[to],1);
    } else if (ban[from]==EKE && dan>=8) {
      // 必ず成る
      teTop[teNum++]=Te(from,to,ban[from],ban[to],1);
    } else if ((ban[from]==EFU || ban[from]==EKY) && dan>=9) {
      // 必ず成る
      teTop[teNum++]=Te(from,to,ban[from],ban[to],1);
    } else {
      if (SorE==SELF && (fromDan<=3 || dan <=3) && CanPromote[ban[from]]) {
	teTop[teNum++]=Te(from,to,ban[from],ban[to],1);
      } else if (SorE==ENEMY && (fromDan>=7 || dan>=7) && CanPromote[ban[from]]){
	teTop[teNum++]=Te(from,to,ban[from],ban[to],1);
      }
      // 成らない手も生成する。
      teTop[teNum++]=Te(from,to,ban[from],ban[to],0);
    }
  }
}


/** 飛車角香車がまっすぐに進む手の生成.
    AddMoves() だけから呼び出される。
 */
void Kyokumen::AddStraight(int SorE, int& teNum, Te* teTop,int from,
			   int dir,int pin,int Rpin)
{
  if (dir==Rpin || dir==-Rpin) {
    return;
  }
  int i;
  if (pin == 0 || pin == dir || pin == -dir) {
    // 空白の間、動く手を生成する
    for (i = dir; ban[from + i] == EMPTY; i += dir)
      AddMove(SorE,teNum,teTop,from,i,0);

    // 味方の駒でないなら、そこへ動く
    if (!(ban[from + i] & SorE)) {
      AddMove(SorE, teNum, teTop, from, i, 0);
    }
  }
}


//toに動く手の生成
void Kyokumen::MoveTo(int SorE, int& teNum, Te *teTop,int to,int* pin)
{
  int p;
  KomaInf koma;

  for (int i = 0; i < DIRECTION_SIZE; i++) {
    if ((koma = ban[to-Direct[i]]) == EMPTY) {
      p = search(to,-Direct[i]);
      if ((ban[p] & SorE) && CanJump[i][ban[p]]) {
	AddMove(SorE,teNum,teTop,p,to-p,pin[p]);
      }
    } else {
      if ((koma & ~SorE) != OU && (koma & SorE) && (CanMove[i][koma]||CanJump[i][koma])) {
	AddMove(SorE, teNum, teTop, to - Direct[i], Direct[i], pin[to-Direct[i]]);
      }
    }
  }
}


static const char* numstr[] = {
  "一", "二", "三", "四", "五", "六", "七", "八", "九", 
  "10", "11", "12", "13", "14", "15", "16", "17", "18"
};

// それっぽく表示する。
void Kyokumen::FPrint(FILE* fp) const
{
  int x, y;
  y = 0;
  fprintf(fp, "Hash:%16llx Hand:%16llx Kyokumen:%16llx\n",
	  HashVal,HandHashVal,KyokumenHashVal);

  fprintf(fp,"持ち駒：");
  for (x = EHI; x >=EFU; x--) {
    if (Hand[x] > 1) {
      y = 1;
      fprintf(fp,"%s%s", komaStr2[x], numstr[Hand[x] - 1]);
    } 
    else if (Hand[x] == 1) {
      y = 1;
      fprintf(fp,"%s", komaStr2[x]);
    }
  }
  if (y) {
    fprintf(fp,"\n");
  } else {
    fprintf(fp,"なし\n");
  }
  fprintf(fp,"  ９ ８ ７ ６ ５ ４ ３ ２ １ \n");
  fprintf(fp,"+---------------------------+\n");
  for(y=1;y<=9;y++) {
    fprintf(fp,"|");
    for(x=9;x>=1;x--) {
      fprintf(fp,komaStr[ban[x*16+y]]);
    }
    fprintf(fp,"|%s", numstr[y - 1]);
    fprintf(fp,"\n");
  }
  fprintf(fp,"+---------------------------+\n");
  fprintf(fp,"持ち駒：");
  y = 0;
  for (x = SHI; x >= SFU; x--) {
    if (Hand[x] > 1) {
      y = 1;
      fprintf(fp,"%s%s", komaStr2[x], numstr[Hand[x] - 1]);
    } 
    else if (Hand[x] == 1) {
      y = 1;
      fprintf(fp,"%s", komaStr2[x]);
    }
  }
  if (y) {
    fprintf(fp,"\n");
  } else {
    fprintf(fp,"なし\n");
  }
}


const char* komaStrForDump[] = {
  "EMP","   ","   ","   ","   ","   ","   ","   ","   ","   ","   ","   ","   ","   ","   ","   ",
"   ","SFU","SKY","SKE","SGI","SKI","SKA","SHI","SOU","STO","SNY","SNK","SNG","SKI","SUM","SRY",
"   ","EFU","EKY","EKE","EGI","EKI","EKA","EHI","EOU","ETO","ENY","ENK","ENG","EKI","EUM","ERY",
" 壁","   ","   ","   ","   ","   ","   ","   ","   ","   ","   ","   ","   ","   ","   ","   ",
};


void Kyokumen::Dump() const
{
  int x,y;
  printf("TestBan[9][9]={\n");
  for(y=1;y<=9;y++) {
    printf("\t{");
    for(x=9;x>=1;x--) {
      printf("%s",komaStrForDump[ban[x*16+y]]);
      if (x>1) printf(",");
    }
    printf("}");
    if (y<9) printf(",");
    printf("\n");
  }
  printf("};\n");
  printf("Motigoma[EHI+1]={");
  for(int koma=0;koma<=EHI;koma++) {
    printf("%d,",Hand[koma]);
  }
  printf("};\n");
}


// 第３章で追加。交換値を求める。
int kyori(int p1, int p2)
{
  return max(abs(p1/16 - p2/16), abs((p1 & 0x0f) - (p2 & 0x0f)));
}


int Kyokumen::IsCorrectMove(Te& te)
{
  // 第5章で追加。駒打ちの場合に、正しい手かどうかチェックする。
  if (te.from == 0) {
    if (ban[te.to] != EMPTY) return 0;

    if (te.koma == SFU) {
      // 二歩と打ち歩詰めのチェック
      for(int dan=1;dan<=9;dan++) {
	if (ban[(te.to&0xf0)+dan]==SFU) return 0;
      }
      if (te.to==kingE+1) {
	int pin[16*11];
	MakePinInf(pin);
	if (Utifudume(SELF,te.to,pin)) return 0;
      }
    }
    else if (te.koma == EFU) {
      // 二歩と打ち歩詰めのチェック
      for(int dan=1;dan<=9;dan++) {
	if (ban[(te.to&0xf0)+dan]==EFU) return 0;
      }
      if (te.to==kingS-1) {
	int pin[16*11];
	MakePinInf(pin);
	if (Utifudume(ENEMY,te.to,pin)) return 0;
      }
    }
    return 1;
  }

  if (ban[te.from] == SOU) {
    if (controlE[te.to]!=0) {
      return 0;
    } else {
      te.capture=ban[te.to];
      return 1;
    }
  } 
  else if (ban[te.from] == EOU) {
    if (controlS[te.to]!=0) {
      return 0;
    } else {
      te.capture=ban[te.to];
      return 1;
    }
  }
  else if (ban[te.from] == SKE || ban[te.from] == EKE) {
    te.capture=ban[te.to];
    return 1;
  }

  int d = kyori(te.from, te.to);
  if (d==0) return 0;
  int dir=(te.to-te.from)/d;
  if (d==1) {
    te.capture=ban[te.to];
    return 1;
  }

  // ジャンプなので、途中に邪魔な駒がいないかどうかチェックする
  for(int i = 1, pos = te.from + dir; i < d; i++, pos += dir) {
    if (ban[pos] != EMPTY)
      return 0;
  }
  te.capture=ban[te.to];
  return 1;
}


/** 手が、現在の局面で合法か.
    キラー手 killer move などの活用 */
int Kyokumen::IsLegalMove(int SorE, Te& te)
{
  if (!(te.koma & SorE)) {
    // 自分の駒でない駒を動かしている
    return 0;
  }
  if (te.from < 0x11) {
    if (Hand[te.koma]==0) return 0;
  } else {
    if (ban[te.from]!=te.koma) return 0;
  }
  if (ban[te.to] & SorE) {
    // 自分の駒を取っている
    return 0;
  }

  if (IsCorrectMove(te)) {
    // 自玉に王手をかけていないか、実際に動かして調べる
    Kyokumen kk(*this);
    kk.Move(SorE,te);
    if (SorE==SELF && kk.controlE[kk.kingS]) {
      return 0;
    }
    if (SorE==ENEMY && kk.controlS[kk.kingE]) {
      return 0;
    }
    return 1;
  }
  return 0;
}


/** 王手だけを生成する。
 */
int Kyokumen::MakeChecks(int SorE, Te* teBuf, int* pin)
{
  Kyokumen kk(*this);			// 実際に動かしてみる局面
  Kiki* selfControl;	// 手番側から見て、自分の利き
  int enemyKing;				// 手番側から見て、相手の玉の位置

  if (SorE==SELF) {
    selfControl=kk.controlS;
    enemyKing=kingE;
  } else {
    selfControl=kk.controlE;
    enemyKing=kingS;
  }
  int teNum = MakeLegalMoves(SorE, teBuf, pin);
  // 実際に動かしてみて、王手だけを残す。
  // ISSUE: すべての手を生成する必要はまったくない。
  int outeNum = 0;
  for(int i=0;i<teNum;i++) {
    kk=*this;
    kk.Move(SorE,teBuf[i]);
    // 相手の玉に利きが付いていれば王手。
    if (selfControl[enemyKing]) {
      teBuf[outeNum++]=teBuf[i];
    }
  }
  return outeNum;
}

