
#ifndef LES_MOVE_LIST_H
#define LES_MOVE_LIST_H 1

#include "koma_inf.h"
#include <stdio.h>

class Kyokumen;

// 手のクラス
class Te
{
public:
	// どこから・どこへはそれぞれ１Byteであらわせます。
	// 詳しくは局面クラスを参照して下さい。
  unsigned char from, to;

  // 動かした駒
  KomaInf koma;

  // 成/不成り
  unsigned char promote;

  // 取った駒
  KomaInf capture;

  // これは、手の生成の際に種別を用いたい時に使います（将来の拡張用）
  unsigned char Kind;
  // その手の仮評価（手の価値）です
  short value;

public:
  inline bool IsNull() const {
    return from == 0 && to == 0;
  }

  // Teの配列を宣言することが多いので、空のコンストラクタを用意します。
  // こうすることで、Teの配列を宣言するたびに無駄な初期化処理が走ることを避けられます。
  inline Te() {}

  // 本当にTeを空のデータで初期化したい時のためのコンストラクタです。
  inline Te(int i) {
		from=to=koma=capture=promote=Kind=0;
		value=0;
	}

  inline Te(int f, int t, KomaInf k, KomaInf c, int p = 0, int K = 0, int v = 0) {
		from=f;
		to=t;
		koma=k;
		capture=c;
		promote=p;
		Kind=K;
		value=v;
	}

  Te(int SorE, unsigned char f, unsigned char t, const Kyokumen& k);

  // 手を表示したい時に使います。
  void Print() const {
    FPrint(stdout);
  }

  // 同上
  void FPrint(FILE* fp) const;

	// 手の同一性を比較したいときに使います。KindやValueが違っても同じ手です。
	int operator==(const Te &a) {
		return a.from==from && a.to==to && a.koma==koma && a.promote==promote;
	}
};



#endif
